/*  edump -- capture data from network interfaces
 *   
 *  Copyright (C) 1996, 1997 Peter Tobias  <tobias@et-inf.fho-emden.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *	   
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *	       
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>		/* atoi */
#include <string.h>		/* rindex */
#include <signal.h>		/* signal */
#include <getopt.h>		/* GNU getopt */
#include <sys/socket.h>		/* inet_ntoa */
#include <netinet/in.h>		/* inet_ntoa */
#include <arpa/inet.h>		/* inet_ntoa */
#include <locale.h>		/* setlocale */
#include <time.h>		/* time/localtime/strftime */
#include <pcap.h>
#include "edump.h"
#include "extern.h"
#include "privs.h"

#ifdef __GLIBC__
# ifndef HAVE_SNPRINTF
#  define HAVE_SNPRINTF
# endif
#endif


#define OPTIMIZER	1	/* run packet-matching code optimizer */
#define NO_OPT		-1
#define SNAPLEN		1514	/* libpcap now uses dynamic frame length
				   so we can use a high value */

static const char *version_string = "edump 1.3";

char *prog_name;
static pcap_t *pd;

void sigexit(int sig)
{
    struct pcap_stat pstat;

    /* only print statistics if we really have a valid
     * packet capture descriptor
     */
    if (pd != NULL) {
	if (pcap_stats(pd, &pstat) >= 0) {
	    fprintf(stderr, "\nFrames received by filter  : %u\n",
	      pstat.ps_recv);
	    fprintf(stderr, "Frames dropped by kernel   : %u\n",
	      pstat.ps_drop);
#ifdef __linux__
	    fprintf(stderr, "Frames dropped by interface: %u\n",
	      pstat.ps_ifdrop);
#endif
	} else {
	    fprintf(stderr, "%s: pcap_stats: %s\n", prog_name, pcap_geterr(pd));
	}
	fflush(stderr);
    }
    /* turn back to root privileges to allow the
     * program to restore the interface
     */
    PRIV_ROOT;
    exit(0);
}

void sigstat(int sig)
{
    struct pcap_stat pstat;
    struct tm *tm;
    time_t tp;
    char t_date[64];
    char t_time[64];

    signal (sig, sigstat);

    /* only print statistics if we really have a
     * valid packet capture descriptor
     */
    if (pd != NULL) {
	tp = time(NULL);
	if (pcap_stats(pd, &pstat) >= 0) {
	    tm = localtime(&tp);
	    strftime(t_date, 63, "%x", tm);
	    strftime(t_time, 63, "%X %Z", tm);
	    fprintf(stderr, "\n- Current date               : %s\n", t_date);
	    fprintf(stderr, "- Current time               : %s\n", t_time);
	    fprintf(stderr, "- Frames received by filter  : %u\n",
	      pstat.ps_recv);
	    fprintf(stderr, "- Frames dropped by kernel   : %u\n",
	      pstat.ps_drop);
#ifdef __linux__
	    fprintf(stderr, "- Frames dropped by interface: %u\n",
	      pstat.ps_ifdrop);
#endif
	} else {
	    fprintf(stderr, "%s: pcap_stats: %s\n", prog_name, pcap_geterr(pd));
	}
	fflush(stderr);
    }
}

void usage(void)
{
    printf("Usage: %s [OPTIONS]... [filter rules]\n\n", prog_name);
    printf("  -n, --no-promisc               don't use the promiscuous mode\n");
    printf("  -s, --snaplen size             set the snaplength"
      " (default 1514 bytes)\n");
    printf("  -c, --count count              limit number of frames to capture\n");
    printf("  -t, --timeout timout           set the read timeout"
      " (default 1000 ms)\n");
    printf("  -i, --interface iface          set the interface to"
      " capture from\n");
    printf("  -f, --filter-file filename     read filters from file\n");
    printf("  -r, --read-file filename       read data from file\n");
    printf("  -w, --write-file filename      set name of capture file\n");
    printf("  -d, --date-name                add date string to filename\n");
    printf("  -W, --wait-dev                 wait for device if not available\n");
    printf("  -h, --help                     display this help and exit\n");
    printf("  -v, --version                  display version information"
      " and exit\n");
    printf("\nYou can specify a filter rule on the command line"
      " (e.g. \"arp or rarp\").\n");
    printf("Filters set on the command line override filters read from file!\n");
}

int main(int argc, char **argv)
{
    struct bpf_program filtercode;
    struct pcap_stat pstat;
    bpf_u_int32 netmask, network=0;
    u_char *pcap_userdata;
    pcap_dumper_t *p;
    char *device;
    char filtercmd[1024]="";
    char ffile[FILENAME_MAX];
    char rfile[FILENAME_MAX];
    char wfile[FILENAME_MAX];
    char ebuf[PCAP_ERRBUF_SIZE];
    char datestr[256];
    time_t timep;
    struct tm *tm;

    int option_index = 0;
    int option_c;
    int waitdev = 0;		/* wait for device */
    int counter = -1;		/* process all frames */
    int promisc = 1;		/* use promiscous mode on interface */
    int timeout = 1000;		/* read timeout in ms */
    int snaplen = SNAPLEN;

    static const struct option long_options[] = {
	{"no-promisc", no_argument, 0, 'n'},
	{"snaplen", required_argument, 0, 's'},
	{"count", required_argument, 0, 'c'},
	{"timeout", required_argument, 0, 't'},
	{"interface", required_argument, 0, 'i'},
	{"filter-file", required_argument, 0, 'f'},
	{"read-file", required_argument, 0, 'r'},
	{"write-file", required_argument, 0, 'w'},
	{"date-name", no_argument, 0, 'd'},
	{"wait-dev", no_argument, 0, 'W'},
	{"help", no_argument, 0, 'h'},
	{"version", no_argument, 0, 'v'},
	{0, 0, 0, 0}
    };
    
    prog_name = (rindex(argv[0], '/')) ? rindex(argv[0], '/') + 1 : argv[0];

    if (geteuid() != 0) {
	printf("%s requires root privileges. See %s(8) for\n", prog_name, prog_name);
	printf("information on how to run it as a non root user.\n");
	exit(0);
    }

    /* drop root privileges */
    U_DROP_PRIVS;

    device = NULL;		/* no default device */
    datestr[0] = '\0';		/* no default date string in filename by default */
    ffile[0] = '\0';		/* no default filter filename */
    rfile[0] = '\0';		/* no default read filename */
    wfile[0] = '\0';		/* no default write filename */

    timep = time(NULL);
    tm = localtime(&timep);

    setlocale(LC_ALL, "");	/* allow all locales */

    while (1) {
	option_c = getopt_long(argc, argv, "c:df:hi:nr:s:t:vw:W", long_options,
	  &option_index);

	if (option_c == NO_OPT)
	    break;

	switch(option_c) {
	    case 'c':
		counter = atoi(optarg);
		break;
	    case 'd':
#ifdef HAVE_SNPRINTF
		snprintf(datestr, sizeof(datestr), "--%04d-%02d-%02d--%02d:%02d:%02d",
		  1900 + tm->tm_year, tm->tm_mon, tm->tm_mday, tm->tm_hour,
		  tm->tm_min, tm->tm_sec);
#else
		sprintf(datestr, "--%04d-%02d-%02d--%02d:%02d:%02d",
		  1900 + tm->tm_year, tm->tm_mon, tm->tm_mday, tm->tm_hour,
		  tm->tm_min, tm->tm_sec);
#endif
		break;
	    case 'f':
		strncpy(ffile, optarg, sizeof(ffile));
		break;
	    case 'h':
		usage();
		exit(0);
		break;
	    case 'i':
		device = optarg;
		break;
	    case 'n':
		promisc = 0;
		break;
	    case 'r':
		strncpy(rfile, optarg, sizeof(rfile));
		break;
	    case 's':
		snaplen = atoi(optarg);
		break;
	    case 't':
		timeout = atoi(optarg);
		break;
	    case 'v':
		printf("%s\n", version_string);
		exit(0);
		break;
	    case 'w':
		strncpy(wfile, optarg, sizeof(wfile));
		break;
	    case 'W':
		waitdev = (1 - waitdev);
		break;
	    case '?':
		fprintf(stderr, "Try `%s --help' for more information.\n", prog_name);
		exit(1);
		break;
	    default:
		break;
	}
    }

    if (strlen(ffile) > 0) {
	register FILE *fd;
	register char *p;
	char fline[1024];

	if ((fd = fopen(ffile, "r")) != NULL) {
	    while (fgets(fline, sizeof(fline), fd) != NULL) {
		if ((p = index(fline, '\n')) != NULL)
		    *p = '\0';
		if (fline[0] == '#' || fline[0] == '\0')
		    continue;
		if (filtercmd[0] != '\0')
		    strcat(filtercmd, " ");
		strcat(filtercmd, fline);
	    }
	    fclose(fd);
	} else {
	    fprintf(stderr, "%s: can't open `%s'\n", prog_name, ffile);
	    exit(1);
	}
    }

    if (optind < argc) {
	if (strlen(ffile) > 0)
	    warn("Warning! Overriding filters read from file!");
	while (optind < argc) {
	    strcat(filtercmd, argv[optind++]);
	    if (optind != argc)
		strcat(filtercmd, " ");
	}
    }

    if (strlen(rfile) > 0) {
	pd = pcap_open_offline(rfile, ebuf);
	if (pd == NULL)
	    error(ebuf);
	network=0;
	netmask=0;
	if (strlen(ffile) > 0)
	    error("can't use the -f option with the -r option");
    } else {
	/* libpcap < 0.2: without this one pcap_lookupnet won't work */
	fflush(stdout);

	if (device == NULL) {
	    /* no device given, try to find one */
	    device = pcap_lookupdev(ebuf);
	    if (device == NULL)
		error(ebuf);
	}

	/* start of block which uses root privileges
	 * (necessary to create socket)
	 */
	PRIV_ROOT_START;

	/* get a "live" packet capture descriptor */
	if ((pd = pcap_open_live(device, snaplen, promisc, timeout, ebuf)) ==
	  NULL)
	    error(ebuf);

	/* end of block which uses root privileges */
	PRIV_ROOT_END;

	/* get network number and netmask of current device */
	if (waitdev) {
	    fprintf(stderr, "%s: waiting for device %s"
	      " (currently not available)\n\n", prog_name, device);
	    /* network == 0x1000000 work around a bug in linux */
	    while ((pcap_lookupnet(device, &network, &netmask, ebuf) < 0) ||
	      (network == 0) || (network == 0x1000000)) {
		sleep(1);	/* do nothing */
	    }
	} else {
	    if (pcap_lookupnet(device, &network, &netmask, ebuf) < 0)
		error(ebuf);
	}
    }

    if (strlen(wfile) == 0) {
	/* buffer overruns in suid programs are dangerous */
	if (strlen(rfile) > 0) {
	    snprintf(wfile, sizeof(wfile), "%s.%s", CAPTUREFILE, "file");
	} else {
	    snprintf(wfile, sizeof(wfile), "%s.%s", CAPTUREFILE, device);
	}
    }

    if (strlen(datestr) > 0) {
	strncat(wfile, datestr, sizeof(wfile));
    }

    if ((pd != NULL) && (strlen(rfile) == 0)) {
	fprintf(stderr, "Device    : %s (promiscuous mode %sabled)\n",
	  device, promisc ? "en" : "dis");
	fprintf(stderr, "Link layer: %s\n", ll_type[pcap_datalink(pd)]);
	fprintf(stderr, "Network   : %s\n",
	  inet_ntoa(*(struct in_addr *)&network));
	fprintf(stderr, "Netmask   : %s\n",
	  inet_ntoa(*(struct in_addr *)&netmask));
	fprintf(stderr, "Snaplen   : %d\n", snaplen);
	fprintf(stderr, "Timeout   : %d\n", timeout);
	fprintf(stderr, "Counter   : %s", (counter <= 0) ? "unlimited\n" : "");
	if (counter > 0)
	    fprintf(stderr, "%u\n", counter);
	fprintf(stderr, "Filter    : %s\n", filtercmd);
	fprintf(stderr, "Filterfile: %s\n", strlen(ffile) ? ffile : "");
	fprintf(stderr, "Readfile  : %s\n", strlen(rfile) ? rfile : "");
	fprintf(stderr, "Savefile  : %s\n\n", wfile);
    }

    /*
     * now that we have a packet capture descriptor and we're going to
     * change some things make sure that there is a clean way out
     */
    signal(SIGHUP, sigexit);
    signal(SIGINT, sigexit);
    signal(SIGTERM, sigexit);

    signal(SIGUSR1, sigstat);

    /* compile string "filtercmd" to a filter program */
    if (pcap_compile(pd, &filtercode, filtercmd, OPTIMIZER, netmask) < 0)
      error(pcap_geterr(pd));

    /* set filter program for packet capture descriptor */
    if (pcap_setfilter(pd, &filtercode) < 0)
	error(pcap_geterr(pd));

    /* prepare capture savefile */
    if ((p = pcap_dump_open(pd, wfile)) == NULL)
	error(pcap_geterr(pd));
    pcap_userdata = (u_char *)p;

    if ((pd != NULL) && (strlen(rfile) == 0)) {
	fprintf(stderr, "%s: monitoring device %s\n", prog_name, device);
    }
    fflush(stderr);

    if (pcap_loop(pd, counter, pcap_dump, pcap_userdata) < 0)
	error(pcap_geterr(pd));

    /* only print statistics if we really have a valid packet
     * capture descriptor
     */
    if ((pd != NULL) && (strlen(rfile) == 0)) {
	if (pcap_stats(pd, &pstat) >= 0) {
	    fprintf(stderr, "\nPackets received by filter  : %u\n",
	      pstat.ps_recv);
	    fprintf(stderr, "Packets dropped by kernel   : %u\n",
	      pstat.ps_drop);
#ifdef __linux__
	    fprintf(stderr, "Packets dropped by interface: %u\n",
	      pstat.ps_ifdrop);
#endif
	} else {
	    fprintf(stderr, "%s: pcap_stats: %s\n", prog_name, pcap_geterr(pd));
	}
	fflush(stderr);
    }

    pcap_close(pd);

    /* turn back to root privileges to allow the program
     * to restore the interface
     */
    PRIV_ROOT;
    return(0);
}

